#version 460 core
layout(location = 0) out vec4 FragColor;
layout(location = 1) out vec4 BrightColor;

struct DirLight {
	vec3 direction;

	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};

struct PointLight {
	vec3 position;

	float constant;
	float linear;
	float quadratic;

	vec3 ambient;
	vec3 diffuse;
	vec3 specular;

	bool isActive;
};

struct SpotLight {
	vec3 position;
	vec3 direction;
	float cutOff;
	float outerCutOff;

	float constant;
	float linear;
	float quadratic;

	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};

struct OmniShadowMap
{
	samplerCube shadowMap;
	float farPlane;
};

// Imports the current position from the Vertex Shader
//FragPos
in vec3 crntPos;
// Imports the normal from the Vertex Shader
in vec3 Normal;
// Imports the texture coordinates from the Vertex Shader
//TexCoords
in vec2 texCoord;
// Imports the position of the directional light from the Vertex Shader
in vec4 DirectionalLightSpacePos;

#define MAX_DIR_LIGHTS 1
#define MAX_POINT_LIGHTS 9
#define MAX_SPOT_LIGHTS 1

uniform int numberDirLights;
uniform int numberPointLights;
uniform int numberSpotLights;

uniform PointLight pointLights[MAX_POINT_LIGHTS];
uniform DirLight dirLights[MAX_DIR_LIGHTS];
uniform SpotLight spotLights[MAX_SPOT_LIGHTS];

uniform vec3 viewPos;

uniform sampler2D texture_diffuse0;
uniform sampler2D texture_specular0;
uniform sampler2D texture_normal0;

uniform sampler2D directionalShadowMap;
uniform OmniShadowMap omniShadowMaps[MAX_POINT_LIGHTS];

vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1), 
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);

float CalcDirectionalShadowFactor(DirLight light)
{
	vec3 normal = normalize(Normal);
	vec3 lightDir = normalize(light.direction - crntPos);

	// perform perspective divide
    vec3 projCoords = DirectionalLightSpacePos.xyz / DirectionalLightSpacePos.w;

    // transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;

    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    float closestDepth = texture(directionalShadowMap, projCoords.xy).r;

    // get depth of current fragment from light's perspective
    float currentDepth = projCoords.z;
	
	//float bias = max(0.05 * (1.0 - dot(normal, lightDir)), 0.0005);
	float bias = min(max(0.001 * (1.0 - dot(normal, lightDir)), 0.0005), 0.001);

	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(directionalShadowMap, 0);
	for(int x = -1; x <= 1; ++x)
	{
		for(int y = -1; y <= 1; ++y)
		{
			float pcfDepth = texture(directionalShadowMap, projCoords.xy + vec2(x, y) * texelSize).r;
			shadow += currentDepth - bias > pcfDepth ? 1.0 : 0.0;
		}    
	}
	shadow /= 12.0;

	if(projCoords.z > 1.0)
        shadow = 0.0;
    
    return shadow;
}

float CalcPointShadowFactor(PointLight light, int shadowIndex)
{
	vec3 fragToLight = crntPos - light.position;
	float currentDepth = length(fragToLight);
	
	float shadow = 0.0;
	
	vec3 normal = normalize(Normal);
	float bias = min(max(0.001 * (1.0 - dot(normal, fragToLight)), 0.005), 0.01);

	int samples  = 20; // number influences required performance 
	float viewDistance = length(viewPos - crntPos);
	float diskRadius = (1.0 + (viewDistance / omniShadowMaps[shadowIndex].farPlane)) / 120.0;
	for(int i = 0; i < samples; ++i)
	{
		float closestDepth = texture(omniShadowMaps[shadowIndex].shadowMap, fragToLight + gridSamplingDisk[i] * diskRadius).r;
		closestDepth *= omniShadowMaps[shadowIndex].farPlane;
		if(currentDepth - bias > closestDepth)
			shadow += 1.0;
	}
	shadow /= float(samples);  
	
	return shadow;
}

vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir)
{
	float shadowFactor = CalcDirectionalShadowFactor(light);

	vec3 lightDir = normalize(light.direction - crntPos);
	// diffuse shading
	float diff = max(dot(normalize(normal), normalize(lightDir)), 0.0);

	// specular shading
	vec3 reflectDir = reflect(-lightDir, normal);
	float spec = pow(max(dot(viewDir, reflectDir), 0.0), 8.0);

	// combine results
	vec3 ambient = light.ambient * vec3(texture(texture_diffuse0, texCoord) * 5.0f);
	vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse0, texCoord) * 5.0f);
	vec3 specular = light.specular * spec * vec3(texture(texture_specular0, texCoord));

	return (ambient + ((1.0f - shadowFactor) * (diffuse)));
}

vec3 CalcPointLight(PointLight light, int shadowIndex, vec3 normal, vec3 crntPos, vec3 viewDir)
{
	float shadowFactor = CalcPointShadowFactor(light, shadowIndex);

	vec3 lightDir = normalize(light.position - crntPos);
	// diffuse shading
	float diff = max(dot(normal, lightDir), 0.0);

	// specular shading
	vec3 reflectDir = reflect(-lightDir, normal);
	float spec = pow(max(dot(viewDir, reflectDir), 0.0), 8.0);

	// attenuation
	float distance = length(light.position - crntPos);
	float attenuation = 1.0 / (light.constant + light.linear * distance +light.quadratic * (distance * distance));

	// combine results
	vec3 ambient = light.ambient * vec3(texture(texture_diffuse0, texCoord));
	vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse0, texCoord));
	vec3 specular = light.specular * spec * vec3(texture(texture_specular0, texCoord));
	ambient *= attenuation;
	diffuse *= attenuation;
	specular *= attenuation;

	return (ambient + ((1.0f - shadowFactor) * (diffuse)));
}

vec3 CalcSpotLight(SpotLight light, vec3 normal, vec3 crntPos, vec3 viewDir)
{
	vec3 lightDir = normalize(light.position - crntPos);
	// diffuse shading
	float diff = max(dot(normal, lightDir), 0.0);

	// specular shading
	vec3 reflectDir = reflect(-lightDir, normal);
	float spec = pow(max(dot(viewDir, reflectDir), 0.0), 8.0);

	// attenuation
	float distance = length(light.position - crntPos);
	float attenuation = 1.0 / (light.constant + light.linear * distance + light.quadratic * (distance * distance));
	
	// spotlight intensity
	float theta = dot(lightDir, normalize(-light.direction));
	float epsilon = light.cutOff - light.outerCutOff;
	float intensity = clamp((theta - light.outerCutOff) / epsilon, 0.0, 1.0);

	// combine results
	vec3 ambient = light.ambient * vec3(texture(texture_diffuse0, texCoord));
	vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse0, texCoord));
	vec3 specular = light.specular * spec * vec3(texture(texture_specular0, texCoord));
	ambient *= attenuation * intensity;
	diffuse *= attenuation * intensity;
	specular *= attenuation * intensity;

	return (ambient + diffuse);
}

void main()
{
	vec3 norm = normalize(Normal);
	vec3 viewDir = normalize(viewPos - crntPos);
	vec3 finalColour = vec3(0.0, 0.0, 0.0);
	
	// phase 1: Directional lighting	
	for (int i = 0; i < numberDirLights; i++){
		finalColour += CalcDirLight(dirLights[i], norm, viewDir);
	}
	
	// phase 2: Point lights
	for (int i = 0; i < numberPointLights; i++)
	{
		if (pointLights[i].isActive)
			finalColour += CalcPointLight(pointLights[i], i, norm, crntPos, viewDir);
	}
			
	// phase 3: Spot lights
	for (int i = 0; i < numberSpotLights; i++){
		finalColour += CalcSpotLight(spotLights[i], norm, crntPos, viewDir);
	}

	FragColor = vec4(finalColour,1.0f);

	float brightness = dot(finalColour, vec3(0.2126, 0.7152, 0.0722));
	if (brightness > 1.5) {
		BrightColor = vec4(finalColour, 1.0);
	}
	else {
		BrightColor = vec4(0.0, 0.0, 0.0, 1.0);
	}
}
